/*
/*   Copyright (c) 2003, David S. Madole
/*   All rights reserved.
/*
/*   For latest sources and information on this software, please
/*   go to http://www.madole.net/shaper
/*
/*   Questions, comments or suggestions should be directed to
/*   David S. Madole <david@madole.net>.
/*
/*   Redistribution and use in source and binary forms, with or without
/*   modification, are permitted provided that the following conditions
/*   are met:
/*
/*   1. Redistributions of source code must retain the above copyright
/*      notice unmodified, this list of conditions, and the following
/*      disclaimer.
/*
/*   2. Redistributions in binary form must reproduce the above copyright
/*      notice, this list of conditions and the following disclaimer in the
/*      documentation and/or other materials provided with the distribution.
/*
/*   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
/*   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
/*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
/*   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
/*   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
/*   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
/*   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
/*   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
/*   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
/*   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
/*   SUCH DAMAGE.
/*
/*   Version 0.9.2 - 05/27/2003
/*   Version 0.9.5 - 01/28/2005
/*   Version 0.9.6 - 11/13/2005
*/


#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#define IPHEADER_FO 0x1FFF
#define IPHEADER_MF 0x2000
#define IPHEADER_DF 0x4000


#define URG 0x20
#define ACK 0x10
#define PSH 0x08
#define RST 0x04
#define SYN 0x02
#define FIN 0x01


struct tcpheader
{
    unsigned short src, dst;
    unsigned long seq, ack;
    unsigned char off, flag;
    unsigned short win, sum, urg;
    unsigned char opts[0]; } ;


struct ipheader
{
    unsigned char hlen, tos;
    unsigned short tlen, ident, off;
    unsigned char ttl, prot;
    unsigned short sum;
    unsigned long src, dst;
    unsigned char opts[0]; } ;


struct icmpheader
{
    unsigned char type, code;
    unsigned short sum;
    unsigned short unused, mtu;
    unsigned char data[0]; } ;


struct ipentry
{
    struct ipentry *next;
    struct sockaddr intf[1];
    unsigned long size;
    unsigned long bits;
    struct ipheader ip[0]; } ;


struct ipqueue
{
    struct ipqueue *next;
    struct ipentry *head;
    struct ipentry *tail;
    struct ippipe *pipe;
    long bits;
    long delay;
    int socket;
    int mtu;
    int mss;
    int wobble;
    int ignoredf;
    int port; } ;


enum ipencaps
{
    ETHER,
    ATM,
    PPP,
    END } ;


struct ippipe
{
    struct ippipe *next;
    struct ipqueue *head;
    struct ipqueue *tail;
    long speed;
    long preload;
    long pending;
    int mtu;
    int mss;
    int wobble;
    int ignoredf;
    enum ipencaps encaps[10]; } ;


/*  Allocate a small number of global variables. The time of packet arrival
/*  in "now" is used outside of the main loop as a random number source. */

struct timeval now[1];

struct ipentry *ipfree = NULL;

int maxmtu = 0;


char *
inet_ltoa(unsigned long addr)
{
    struct in_addr inet;

    inet.s_addr = addr;
    return inet_ntoa(inet);
}


void
dumppacket(struct ipheader *ip)
{
    unsigned char *p;
    int count;
    int len;

    len = ntohs(ip->tlen);
    p = (char *)ip;
    count = 0;

    while (len--) {
	if ((count % 16) == 0) printf("    %04x", count);
	printf(" %02x", *p++);
	if ((count++ % 16) == 15) printf("\n");
    }
    if ((count++ % 16) != 15) printf("\n");
}


/*  Calculate a standard UP checksum for a data block. If the checksum field
/*  included in the block, it should be set to zero before invoiking this.
/*  I haven't determined yet if this checksum function really works correctly
/*  on odd-length data blocks, so use this elsewhere with caution. In this
/*  program it is only used for ICMP and IP header checksums, which are always
/*  a multiple of four.  */

unsigned short
checksum(void *data, int len)
{
    unsigned long sum;
    unsigned short *word;

    word = (unsigned short *)data;

    for (sum = 0; len > 1; len -= 2) sum += *word++;

    if (len) sum += *word & ntohs(0xff);

    sum = (sum >> 16) + (sum & 0xffff);
    sum = (sum >> 16) + (sum & 0xffff);

    return (sum ^ 0xffff);
}


/*  Incrementaly update a checksum, given old checksum, the old value of the
/*  16 bit word that was changed, and the new value of the changed word. This
/*  is faster if only a small amount of data is begin changed, plus in some
/*  cases, we don't have all the data to recalculate over anyway. */

unsigned short
incrsum16(unsigned short sum, unsigned short old, unsigned short new)
{
    unsigned long tmp;

    tmp = (unsigned short)~sum + htons(~old) + htons(new);

    return ~((tmp >> 16) + (tmp & 0xffff));
}


/*  If we receive a packet that is marked "don't fragment" and it won't
/*  fit in the configured MTU and we aren't ignoring DF, we need to send
/*  an ICMP "fragmentation needed" message in reply. Here we compose and
/*  send such a reply based on the ip packet header that is passed. */

void
sendneedfrag(int sock, struct ipheader *ip, int mtu)
{
    int datlen;
    struct sockaddr_in addr;
    static struct icmpheader *icmp = NULL;


    /*  This should only be sent in response to an initial fragment, and
    /*  not in response to ICMP packets. The latter condition avoids loops.

    if (ntohs(ip->off) & IPHEADER_FO || ip->prot == 1) return;


    /*  The first time we are called, allocate a semi-populated packet that
    /*  we will keep and reuse. For each invocation, populate the data that
    /*  can change from use to use. */

    if (icmp == NULL) {

	icmp = malloc(sizeof(*icmp) + 72);

	if (icmp == NULL) return;

	icmp->type = 3;
	icmp->code = 4;
	icmp->unused = 0;
    }

    icmp->mtu = htons(mtu);
    datlen = (ip->hlen & 0xf) * 4 + 8;

    memcpy(icmp->data, ip, datlen);
    memset(icmp->data + datlen, 0, 72 - datlen);

    icmp->sum = 0;
    icmp->sum = checksum(icmp, sizeof(*icmp) + datlen);


    /*  Send the completed packet out through the raw ICMP socket that was
    /*  allocated when the program was first started. */

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = ip->src;
    addr.sin_port = 0;

    sendto(sock, icmp, sizeof(*icmp) + datlen, 0,
	  (struct sockaddr *) &addr, sizeof(addr));
}




int
queuepacket(struct ipqueue *queue, struct ipheader *ip, struct sockaddr *intf)
{
    struct sockaddr_in *addr;

    int optidx, optlen, optmax, optcpy;
    int hdrlen, pktlen, frglen, datlen, offset;

    enum ipencaps *encaps;
    struct ipentry *entry;


    /*  If the queue is already full, don't queue the packet, but return
    /*  success anyway. The packet will just get dropped. */

    if (queue->bits > queue->delay) return 0;


    /*  Calculate the offset and lengths of the parts of the datagram. If the
    /*  packet ends up being an illegal length, discard it now. Otherwise,
    /*  get a copy of the interface socket structure just to change type. */

    offset = 8 * (ntohs(ip->off) & IPHEADER_FO);
    hdrlen = 4 * (ip->hlen & 0xf);
    datlen = ntohs(ip->tlen) - hdrlen;

    if (datlen < 0 || offset + datlen > 65535) return 0;

    addr = (struct sockaddr_in *)intf;


    /*  Unless this queue is configured to ignore the DF bit, return an error
    /*  if DF is set and the packet is too big to pass in one piece. The
    /*  caller will send an ICMP must-fragment message back if needed. */

    if (queue->ignoredf == 0 &&
	ip->off & htons(IPHEADER_DF) != 0 &&
	hdrlen + datlen > queue->mtu) return -1;


    /*  If MSS override has been configured and the packet carries a TCP
    /*  header, scan through the options and if a MSS option is present,
    /*  update it to the configured value. If we change it, update the TCP
    /*  checksum incrementally since it's faster, plus we may not
    /*  even have all the data if this datagram was fragmented. */

    if (queue->mss > 0 && ip->prot == 6 && offset == 0) {

	struct tcpheader *tcp = (struct tcpheader *)((char *)ip + hdrlen);

	optmax = ((tcp->off >> 4) << 2) - sizeof(*tcp);

	for (optidx = 0; optidx < optmax; ) {

	    switch(tcp->opts[optidx]) {

		unsigned short oldmss;

	    case 0:
		optidx = optmax;
		break;

	    case 1:
		optidx++;
		break;

	    case 2:

		oldmss  = (tcp->opts[optidx + 2] << 8) + tcp->opts[optidx + 3];

		if (oldmss > queue->mss) {

		    tcp->opts[optidx + 2] = queue->mss >> 8;
		    tcp->opts[optidx + 3] = queue->mss & 0xff;

		    tcp->sum = incrsum16(tcp->sum, oldmss, queue->mss);
		}

	    default:
		optidx += tcp->opts[optidx + 1];
		break;
	    }
	}
    }


    /*  Figure out the number of octets of IP options that are present. Then
    /*  copy the packet to one or more queue buffers, fragmenting as needed
    /*  during the process of copying. For each fragment, reuse an existing
    /*  buffer from the free list if one is available, otherwise allocate
    /*  a fresh buffer. */

    optmax = hdrlen - sizeof(*ip);

    while (datlen > 0) {

	if (ipfree == NULL) {
	    entry = malloc(sizeof(*entry) + maxmtu);
	    if (entry == NULL) return 0;
	}
	else {
	    entry = ipfree;
	    ipfree = entry->next;
	}


	/*  Copy any IP options attached to the packet, except note that
	/*  only options that have the high-order bit of the option type set
	/*  should be copied to fragments after the first one. */

	for (optidx = optcpy = 0; optidx < optmax; ) {

	    switch (ip->opts[optidx]) {

	    case 0:
		optidx = optmax;
		break;

	    case 1:
		if (offset == 0)
		    entry->ip->opts[optcpy++] = ip->opts[optidx];
		optidx++;
		break;

	    default:
		optlen = ip->opts[optidx + 1];

		if (optidx + optlen <= optmax &&
		    ip->opts[optidx] & 0x80 || offset == 0) {

		    memcpy(entry->ip->opts + optcpy, ip->opts + optidx, optlen);
		    optcpy += optlen;
		}
		optidx += optlen;
		break;
	    }
	}


	/*  If we are not left pointing exactly at the end of the options,
	/*  then there were bogus options in this packet, that is, the length
	/*  stored in the option was not valid. Discard the packet if this
	/*  happens. Otherwise, pad the options out to a four byte boundary
	/*  with zeroes if needed, to clear whatever may have been there. */

	if (optidx != optmax) {

	    entry->next = ipfree;
	    ipfree = entry;
	    return 0;
	}

	while (optcpy % 4) entry->ip->opts[optcpy++] = 0;


	/*  If the packet is too large for the MTU, we need to fragment it.
	/*  Start by figuring the largest fragment that will fit in the MTU,
	/*  then if "wobble" is enabled, make the fragment randomly smaller
	/*  by up to the amount of "wobble" configured. If DF is set because
	/*  we ignored it earlier, clear it now since it makes doesn't make
	/*  sense to say don't fragment when it's already fragmented. */

	if (sizeof(*ip) + optcpy + datlen > queue->mtu) {

	    frglen = ((queue->mtu - sizeof(*ip) - optcpy) / 8) * 8;

	    if (queue->wobble) {

		int wobble = frglen / 8;

		if (queue->wobble < wobble) wobble = queue->wobble + 1;

		frglen = frglen - (now->tv_usec % wobble) * 8;
	    }

	    ip->off &= htons(~IPHEADER_DF);
	}
	else frglen = datlen;


	/*  Copy the header and data portion of the fragment. The options,
	/*  if any, have already been copied. Save the total size of the
	/*  new packet in the queue entry in both bytes and bits. */

	memcpy(entry->ip, ip, sizeof(*ip));
	memcpy(entry->ip->opts + optcpy, ip->opts + optmax + offset, frglen);

	entry->size = sizeof(*ip) + optcpy + frglen;
	entry->bits = entry->size * 8;


	/*  If any encapsulations have been configured, manipulate the size
	/*  of the packet in bits to reflect what will actually pass out on
	/*  the wire where the rate limiting occurs. Note that the order
	/*  that multiple encapsulations are applied is important, so do
	/*  them in the order that they were defined. */

	for (encaps = queue->pipe->encaps; *encaps != END; encaps++) {

	    switch (*encaps) {

		case PPP:
		    entry->bits += 64;
		    break;

		case ATM:
		    entry->bits = (entry->bits + 488) / 384 * 424;
		    break;

		case ETHER:
		    entry->bits += 160;
		    if (entry->bits < 528) entry->bits = 528;
		    break;
	    }
	}


	/*  Touch up fields in the header that change for each fragment.
	/*  Then, if the packet is not outgoing, recalculate the IP header
	/*  checksum. The kernel will do this for us on outgoing packets. */

	entry->ip->tlen = htons(entry->size);
	entry->ip->hlen = (4 << 4) + (sizeof(*ip) + optcpy) / 4;
	entry->ip->off = htons(offset / 8);

	datlen -= frglen;
	offset += frglen;

	if (datlen > 0 || ip->off & htons(IPHEADER_MF))
	    entry->ip->off |= htons(IPHEADER_MF);

	if (ip->off & htons(IPHEADER_DF))
	    entry->ip->off |= htons(IPHEADER_DF);

	if (addr->sin_addr.s_addr != INADDR_ANY) {

	    entry->ip->sum = 0;
	    entry->ip->sum = checksum(entry->ip, sizeof(*ip) + optcpy);
	}


	/*  Remember the interface information that the packet was received
	/*  with so we can re-inject it later with the same. Then add the
	/*  packet to the end of the queue and update the queue depth. */

	memcpy(entry->intf, intf, sizeof(*intf));

	entry->next = NULL;

	if (queue->head)
	    queue->tail->next = entry;
	else queue->head = entry;

	queue->tail = entry;

	queue->bits += entry->bits;
    }
    return 0;
}


struct ippipe *
allocpipe(int speed)
{
    struct ippipe *pipe;

    pipe = malloc(sizeof(*pipe));

    if (pipe != NULL) {

	pipe->speed = speed;
	pipe->mtu = 1500;
	pipe->mss = 0;
	pipe->preload = 0;
	pipe->wobble = 0;
	pipe->ignoredf = 0;
	pipe->encaps[0] = END;

	pipe->pending = 0;

	pipe->head = NULL;
	pipe->tail = NULL;
    }
    return pipe;
}


struct ipqueue *
allocqueue(struct ippipe *pipe, int port)
{
    struct sockaddr_in addr[1];
    struct ipqueue *queue;
    int result;
    int sock;

    sock = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT);

    if (sock < 0) return NULL;

    addr->sin_len = 6;
    addr->sin_family = AF_INET;
    addr->sin_addr.s_addr = INADDR_ANY;
    addr->sin_port = htons(port);
    bzero(addr->sin_zero, 8);

    result = bind(sock, (struct sockaddr *)addr, sizeof(*addr));

    if (result < 0) {
	close(sock);
	return NULL;
    }

    result = fcntl(sock, F_SETFL, O_NONBLOCK);

    if (result < 0) {
	close(sock);
	return NULL;
    }

    if (queue = malloc(sizeof(*queue))) {

	queue->bits = 0;
	queue->port = port;
	queue->socket = sock;

	queue->pipe = pipe;
	queue->delay = pipe->speed * 100;
	queue->mtu = pipe->mtu;
	queue->mss = pipe->mss;
	queue->wobble = pipe->wobble;
	queue->ignoredf = pipe->ignoredf;

	queue->next = NULL;
	queue->head = NULL;
	queue->tail = NULL;

	if (pipe->head)
	    pipe->tail->next = queue;
	else pipe->head = queue;

	pipe->tail = queue;
    }
    else close(sock);

    return queue;
}


int main(int argc, char **argv)
{
    int len;
    int size;
    int port;
    int icmp;
    int value;
    int numfd;

    long wakeup;
    long elapsed;
    long earliest;

    fd_set fdset;

    struct ipheader *ip;

    struct sockaddr addr[1];

    struct timeval last[1];
    struct timeval wait[1];
    struct timeval *timeout;

    struct ippipe *list;
    struct ippipe *pipe;
    struct ipqueue *queue;
    struct ipentry *entry;
    enum ipencaps *encaps;

    argv++;

    numfd = 0;
    FD_ZERO(&fdset);

    list = NULL;
    pipe = NULL;


    while (argv[0]) {


	/*  Option: pipe <speed>
	/*
	/*  Creates a new pipe, which represents a set of collectively
	/*  bandwidth-limited queues. The only mandatory parameter of a pipe
	/*  is the bandwidth, which is in kbits per second. */

	if (strcmp(argv[0], "pipe") == 0 && argv[1]) {

	    value = atoi(argv[1]);

	    if (value > 4096) {

		fprintf(stderr, "error: maximum speed is 4096k\n");
		return EXIT_FAILURE;
	    }
	    else if (value < 1) {

		fprintf(stderr, "error: invalid speed setting\n");
		return EXIT_FAILURE;
	    }

	    pipe = allocpipe(value);

	    if (pipe == NULL) {

		fprintf(stderr, "error: can't allocate pipe: %s\n",
		    strerror(errno));
		return EXIT_FAILURE;
	    }

	    pipe->next = list;
	    list = pipe;

	    queue = NULL;
	    encaps = pipe->encaps;

	    argv += 2;
	    continue;
	}


	/*  Option: ethernet
	/*
	/*  Adds per-packet overhead calculation for transmission over
	/*  ethernet to the pipe currently being configured. This accounts
	/*  for MAC-level headers and sync bytes and well as ethernet minimum
	/*  packet length. */

	else if (strcmp(argv[0], "ethernet") == 0) {

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before ethernet\n");
		return EXIT_FAILURE;
	    }
	    *encaps++ = ETHER;
	    *encaps = END;

	    argv += 1;
	    continue;
	}


	/*  Option: atm
	/*
	/*  Adds per-packet overhead calculation for transmission via
	/*  ethernet-to-ATM bridging */

	else if (strcmp(argv[0], "atm") == 0) {

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before ethernet\n");
		return EXIT_FAILURE;
	    }
	    *encaps++ = ATM;
	    *encaps = END;

	    argv += 1;
	    continue;
	}
	else if (strcmp(argv[0], "ppp") == 0) {

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before ppp\n");
		return EXIT_FAILURE;
	    }
	    *encaps++ = PPP;
	    *encaps = END;

	    argv += 1;
	    continue;
	}
	else if (strcmp(argv[0], "mtu") == 0 && argv[1]) {

	    int mtu = atoi(argv[1]);

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before mtu\n");
		return EXIT_FAILURE;
	    }

	    if (mtu < 68) {
		fprintf(stderr, "error: minimum mtu is 68\n");
		return EXIT_FAILURE;
	    }
	    if (queue) queue->mtu = mtu;
	    else pipe->mtu = mtu;

	    if (mtu > maxmtu) maxmtu = mtu;

	    argv += 2;
	    continue;
	}
	else if (strcmp(argv[0], "mss") == 0 && argv[1]) {

	    int mss = atoi(argv[1]);

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before mss\n");
		return EXIT_FAILURE;
	    }
	    if (mss < 48) {
		fprintf(stderr, "error: minimum mss is 48\n");
		return EXIT_FAILURE;
	    }
	    if (queue) queue->mss = mss;
	    else pipe->mss = mss;

	    argv += 2;
	    continue;
	}
	else if (strcmp(argv[0], "ignore-df") == 0) {

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before clear df\n");
		return EXIT_FAILURE;
	    }

	    if (queue) queue->ignoredf = 1;
	    else pipe->ignoredf = 1;

	    argv += 1;
	    continue;
	}
	else if (strcmp(argv[0], "wobble") == 0 && argv[1]) {

	    int wobble = atoi(argv[1]) / 8;

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before mss\n");
		return EXIT_FAILURE;
	    }

	    if (wobble > (queue ? queue->mtu : pipe->mtu) / 8 - 1) {

		fprintf(stderr, "error: maximum wobble is mtu - 8\n");
		return EXIT_FAILURE;
	    }

	    if (queue) queue->wobble = wobble;
	    else pipe->wobble = wobble;

	    argv += 2;
	    continue;
	}
	else if (strcmp(argv[0], "preload") == 0 && argv[1]) {

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before preload\n");
		return EXIT_FAILURE;
	    }
	    pipe->preload = atof(argv[1]) * 1000 * pipe->speed;

	    if (pipe->preload < 0) {
		fprintf(stderr, "error: invalid preload setting\n");
		return EXIT_FAILURE;
	    }
	    argv += 2;
	    continue;
	}
	else if (strcmp(argv[0], "queue") == 0 && argv[1]) {

	    if (pipe == NULL) {
		fprintf(stderr, "error: specify pipe before queue\n");
		return EXIT_FAILURE;
	    }
	    port = atoi(argv[1]);

	    if (port < 1 || port > 65565) {
		fprintf(stderr, "error: invalid port number for queue\n");
		return EXIT_FAILURE;
	    }
	    queue = allocqueue(pipe, port);

	    if (queue == NULL) {
		fprintf(stderr, "setup of port %d failed: %s\n",
			port, strerror(errno));
		return EXIT_FAILURE;
	    }

	    FD_SET(queue->socket, &fdset);

	    if (queue->socket >= numfd)
		numfd = queue->socket + 1;

	    argv += 2;
	    continue;
	}
	else if (strcmp(argv[0], "delay") == 0 && argv[1]) {

	    if (queue == NULL) {
		fprintf(stderr, "error: specify queue before delay\n");
		return EXIT_FAILURE;
	    }
	    queue->delay = atof(argv[1]) * 1000 * pipe->speed;

	    if (queue->delay < 0) {
		fprintf(stderr, "error: invalid preload setting\n");
		return EXIT_FAILURE;
	    }
	    argv += 2;
	    continue;
	}
	if (*argv) {
	    fprintf(stderr, "invalid option '%s'\n", *argv);
	    return EXIT_FAILURE;
	}
    }


    /*  If we didn't end up configuring at least one pipe and one queue,
    /*  output an appropriate error message and give up. */

    if (pipe == NULL) {
	fprintf(stderr, "error: no pipe defined\n");
	return EXIT_FAILURE;
    }
    if (queue == NULL) {
	fprintf(stderr, "error: no queue defined\n");
	return EXIT_FAILURE;
    }


    /*  Allocate a buffer to receive each packet into and open a raw ICMP
    /*  socket to send ICMP fragmentation needed messages though. */

    ip = malloc(65535);

    icmp = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);

    if (icmp < 0) {
	fprintf(stderr, "can't open socket: %s\n", strerror(errno));
	return EXIT_FAILURE;
    }


    /*  Go into the background as a daemon. Fork twice to orphan ourselves,
    /*  change directtory to root, and close all stdio file descriptors. */

    if (fork()) return 0;

    setsid();

    if (fork()) return 0;

    chdir("/");

    close(0);
    close(1);
    close(2);

    timeout = NULL;


    /*  Here is where the business happens. We loop endlessly, using select()
    /*  to wait for packets to arrive based on the FDSET that we already
    /*  built for the ports we listen on. A timeout is used to schedule a
    /*  wakeup for when packets are scheduled to be transmitted.  When we
    /*  do return from select, calculate how long we were sleeping. */

    for ( ; ; ) {

	select(numfd, &fdset, NULL, NULL, timeout);

	gettimeofday(now, NULL);

	elapsed = (now->tv_usec - last->tv_usec)
		+ (now->tv_sec - last->tv_sec) * 1000000;

	last->tv_usec = now->tv_usec;
	last->tv_sec = now->tv_sec;

	earliest = LONG_MAX;


	/*  Now loop through the configured pipes. For each one, if we
	/*  were waiting for any bits to be transmitted, update the pending
	/*  based on how long we were sleeping. */

	for (pipe = list; pipe != NULL; pipe = pipe->next) {

	    if (timeout != NULL && pipe->pending > 0)
		pipe->pending -= elapsed * pipe->speed / 1000;

	    else pipe->pending = 0;


	    /*  Loop through each queue attached to this pipe and check to
	    /*  see if there is a packet waiting. If so, read as many as are
	    /*  ready right now and add them to the queue. If queuepacket
	    /*  returns -1, that means there was an MTU issue, so send a
	    /*  fragment needed response. Finally, make sure the FDSET
	    /*  bit is set again so that we select() on this fd again. */

	    for (queue = pipe->head; queue != NULL; queue = queue->next) {

        	if (FD_ISSET(queue->socket, &fdset)) {

		    for ( ; ; ) {

			size = sizeof(*addr);
			len = recvfrom(queue->socket,
				       ip, 65535, 0, addr, &size);

			if (len < 0) break;

			if (queuepacket(queue, ip, addr) < 0)
			    sendneedfrag(icmp, ip, queue->mtu);
		    }

		    if (errno != EAGAIN)
			syslog(LOG_ERR, "error reading from socket: %s",
			    strerror(errno));

		} else FD_SET(queue->socket, &fdset);


		/*  Next send as many packets from the queue as we can based
		/*  on the number of bits currently pending transmission.
		/*  Since we are iterating through the queues in order and
		/*  sending as much as possible from each before we move on,
		/*  this takes care of prioritization. When we have finally
		/*  filled the pipe, figure out when we need to wakeup
		/*  again and keep track of the earliest wakeup time. */

		for ( ; ; ) {

		    if (pipe->pending > pipe->preload) {

			wakeup = pipe->pending - pipe->preload;
			wakeup = wakeup * 1000 / pipe->speed;

			if (wakeup < earliest) earliest = wakeup;
			break;
		    }

		    if ((entry = queue->head) == NULL) break;

		    len = sendto(queue->socket, entry->ip, entry->size, 0,
				 entry->intf, sizeof(*entry->intf));

		    if (len < 0)
			syslog(LOG_ERR, "error writing to socket: %s",
			    strerror(errno));

		    else pipe->pending += entry->bits;

		    queue->bits -= entry->bits;
		    queue->head = entry->next;

		    entry->next = ipfree;
		    ipfree = entry;
		}
	    }
	}


	/*  If we have sent packets and need to schedule a wakeup time for
	/*  sending more packets from the queue, setup the timeout for the
	/*  next select() loop, Otherwise, select will just wait until we
	/*  receive more packets. */

	if (earliest < LONG_MAX) {

	    wait->tv_sec = earliest / 1000000;
	    wait->tv_usec = earliest % 1000000;

	    timeout = wait;

	} else timeout = NULL;
    }
}

